module Graphics.SVGFonts.Triangulation (adaptTri,Triangle(..),cycle_n) where

import Maybe     ( fromJust, mapMaybe, isJust )
import List      ( elemIndex )
import Array     ( Array, listArray, (!) )
import Graphics.SVGFonts.TriangulationUtils hiding (intersection,vertices)
import Debug.Trace

diagonal    :: (Ord a, Floating a) => [P2 a] -> Line a
diagonal ps = case searchTriangles (m,i,p) ps of
                   Nothing -> Segment m p
                   Just q  -> Segment i q
  where (m,i,p)	= findConvexVertex ps

findConvexVertex              :: (Ord a, Num a) => [P2 a] -> (P2 a, P2 a, P2 a)
findConvexVertex              = head . filter (uncurry3 isLeftTurn) . angles

searchTriangles               :: (Floating a, Ord a) => Angle a -> [P2 a] -> Maybe (P2 a)
searchTriangles x@(m,i,p) ps
  | y == a || y == b          = Just y
  | isJust inLeft  && l /= m  = inLeft
  | isJust inRight && r /= p  = inRight
  | otherwise                 = Nothing
  where ray        = bisectInteriorAngle x
        ps'	       = sublist p m ps
        (a, y, b)  = intersection ray x ps'
        inLeft     = vertexInTriangle (i,y,b) (sublist b m ps')
        (Just l)   = inLeft
        inRight    = vertexInTriangle (y,i,a) (sublist p a ps')
        (Just r)   = inRight

bisectInteriorAngle           :: (Ord a, Floating a) => Angle a -> Line a
bisectInteriorAngle (m,(ix,iy,ii),p)
--  | phi > 0                   = Ray i (rotate m i (-(2*pi-phi)/2))
  | phi > 0                   = Ray i (rotate m (ix,iy,ii) (phi/2 - pi))
--  | phi == 0		      = Ray i (ix+1,iy,ii)
  | phi == 0		      = Ray i (ix+1,iy,ii)
  | phi < 0		      = Ray i (rotate m i (phi/2))
  where phi		      = angle3 p i m
        i = (ix,iy,ii)

intersection                  :: (Ord a, Fractional a) => Line a -> Angle a -> [P2 a] -> Angle a
intersection ray (m,i,p) ps   = minimumWith (\ (_,y,_) -> sqrDistance i y) is
  where 
    is                        = mapMaybe inters (init (edges ps))
    inters (i,j)
      | isJust res            = Just (i,y,j)
      | otherwise             = Nothing
      where
        res                   = intersect ray (Segment i j)
        y                     = fromJust res

vertexInTriangle	      :: (Ord a, Num a) => Angle a -> [P2 a] -> Maybe (P2 a)
vertexInTriangle t@(p,q,r) xs = case filter (Triangle t `contains`) xs of 
				  [] -> Nothing
				  ys -> Just (minimumBy (isLeftTurn p) ys)

-- Wir bentigen folgende Hilfsfunktionen:

-- S.15, |Wedge p q r| ist der Teil der Ebene, der ``rechts'' durch den von |p| ausgehenden Strahl
-- durch |q| und ``rechts'' durch |r| begrenzt wird. Wir bentigen einen Test, ob eine Strecke in
-- einem Keil enthalten ist.

inWedge :: (Ord a, Floating a) => (P2 a, P2 a) -> (P2 a, P2 a, P2 a) -> Bool
(a,b) `inWedge` (p,q,s)
  | r > 0      = not ( (0<i && i<r) || (0<j && j<r) )
  | r < 0      = i<=0 && i>r && j<=0 && j>r
  | r == 0     = True
  | otherwise  = False            --- temp solution
  where r      = angleWrt p q s
        i      = angleWrt p q a
        j      = angleWrt p q b

adaptTri   :: (Floating a, Ord a) => [(a,a)] -> [Triangle a] -- newtype (Num a) => Triangle a = Triangle ((a,a,Int), (a,a,Int), (a,a,Int))
adaptTri poly               = triangulate ds
  where
  ds                        = [(i,j), (j,i)]  -- the polygon is splitted in two parts: the lines from i to j (i,j) and from j to i
    where i                 = sel3_3 xi
          j                 = sel3_3 xj
          (Segment xi xj)   = diagonal ps -- a diagonal is a line inside the polygon

  triangulate []              = []
  triangulate ((i,j):ds)      = map snd ts ++ triangulate (ds' ++ ds)
    where (ds', ts)           = sleeve i j []

  ps                  = vertices poly
  n	                  = length ps
  arr                 = listArray (0,n-1) ps
  pred i              = (i-1) `mod` n
  succ i              = (i+1) `mod` n
  angle i             = (arr!(pred i), arr!i, arr!(succ i))

  sleeve i j ts  -- ts are the accumulated triangles
    | i==j || i==succ j || j==succ i = ([], ts) -- a triangle with i,j and something between
    | (xi,xip) `inWedge` (xi,xj,xim) && (xj,xjm) `inWedge` (xj,xjp,xi) = turns (orientation xjm xj xi) (orientation xip xi xj)
    | otherwise		      = freeTriangle
    where
    (xim,xi,xip)	      = angle i
    (xjm,xj,xjp)	      = angle j
    inBetween	              = sublist xip xjm ps
    index x		      = (i + 1 + (fromJust (elemIndex x inBetween))) `mod` n

    turns Clockwise Clockwise                = nexti
    turns Collinear Clockwise                = nexti
    turns CounterClockwise CounterClockwise	 = prevj
    turns CounterClockwise Collinear		 = prevj
    turns CounterClockwise Clockwise
      | xjm `inInteriourOf` (Triangle (xip,xi,xj)) = prevj
      | otherwise                                  = nexti
    turns _ _
      | null wrongTriangles	     = newFreeTriangle
      | otherwise                = freeTriangle

    nexti      = sleeve (succ i) j (((j,i,succ i), Triangle (xj,xi,xip)) : ts)
    prevj      = sleeve i (pred j) (((j,i,pred j), Triangle (xj,xi,xjm)) : ts)

    (ts', wrongTriangles)	
      | (xi,xip) `inWedge` (xi,xj,xim)  = findWrong ts
      | otherwise = let (vs, w:ss) = splitWhile notInTri ts
                        (xs, ws)   = findWrong ss 
                    in (xs, ws ++ [w] ++ reverse vs)
      where
      notInTri (_,t)          = trace (show (not (t `containsBNV` xip))) (not (t `containsBNV` xip))
      findWrong ts 	          = foldl delete (ts, []) (zip inBetween (tail inBetween))
      delete (ts, ts') (p, q) = (ds, reverse cs ++ ts')
        where
        (cs, ds)              = splitWhile (doIntersect (Segment p q) . diag) ts
        diag ((j,i,n), Triangle (xj,xi,xn)) 
	  | n == succ i       = Segment xn xj
  	  | n == pred j       = Segment xi xn

    -- after Lemma 5
    freeTriangle 
      | null wrongTriangles   = ([], ts)
      | otherwise             = ([(s,k),(k,r)], (undefined, Triangle (xr,xs,xk)) : ts')
      where
      ((r,s,t), tri@(Triangle (xr,xs,xt))) = head wrongTriangles
      inTriangle = filter (tri `containsBNV`) inBetween
      xk         = minimumWith (key xs xr) inTriangle
      k	         = index xk

    key p q x = (angle3 x q p, sqrDistance q x)
   
    -- after Lemma 2+3
    newFreeTriangle	  = ([(i,l),(l,j)], (undefined, Triangle (xj,xi,xl)) : ts)
      where
      xk              = fromJust (searchTriangles (xj,xi,xip) inBetween)
      inTriangle      = filter (Triangle (xj,xi,xk) `contains`)
                               (sublist2 xk xj inBetween)
      xl | null inTriangle = xk
         | otherwise       = minimumWith (key xi xj) inTriangle
      l	              = index xl

-- make vertices of polygon counterclockwise and add an index
vertices :: (Floating a, Ord a) => [(a,a)] -> [(a,a,Int)]
vertices qs | polygon_direction ps = ps
            | otherwise            = reverse ps
  where ps = zipWith (\(x,y) z -> (x,y,z) ) qs [0..]

polygon_direction :: (Floating a, Ord a) => [(a,a,Int)] -> Bool
polygon_direction poly = sum (zipWith crossp l1 l2) > 0
  where l1 = map (\(a,b) -> b `sub` a) ((tail c) ++ [head c])
        l2 = map (\(a,b) -> a `sub` b) c
        c = map to_tuple (cycle_neighbours poly) -- [(p0,p1,0),(p1,p2,1),(p2,p3,2),..
        to_tuple (a:b:[]) = (a,b)
        sub (x0,y0,_) (x1,y1,_) = (x0-x1, y0-y1, 0)
        crossp (v0,v1,_) (w0,w1,_) = v0*w1-v1*w0

-- return a list containing lists of every element with its neighbour
-- i.e. [e1,e2,e3] -> [ [e1,e2], [e2,e3], [e3, e1] ]
cycle_neighbours :: [a] -> [[a]]
cycle_neighbours [] = []
cycle_neighbours xs = cycle_n (head xs) xs

cycle_n :: a -> [a] -> [[a]]
cycle_n f (x:y:xs) = [x,y] : (cycle_n f (y:xs))
cycle_n f e = [[head e, f ]] -- if the upper doesn't match close cycle